from collections import defaultdict
from itertools import chain
from typing import Dict, Optional, Set

from click import Path as ClickPath, secho
from git import Repo
from git.exc import InvalidGitRepositoryError
from tqdm import tqdm

from deckz.cli import command, option
from deckz.exceptions import DeckzException
from deckz.paths import GlobalPaths, Paths
from deckz.targets import Targets


@command
@option(
    "--path",
    type=ClickPath(exists=True, readable=True, file_okay=False),
    default=".",
    help="Path to run the command from.",
)
def deps(path: str) -> None:
    """Give information about shared modules usage."""
    paths = GlobalPaths(path)
    targets_paths = list(paths.git_dir.glob("**/targets.yml"))

    all_targets_dependencies: Dict[str, Dict[str, Set[str]]] = defaultdict(dict)
    for targets_path in tqdm(targets_paths, "Processing targets files"):
        paths = Paths(str(targets_path.parent))
        targets_name = str(paths.current_dir.relative_to(paths.git_dir))
        targets = Targets(paths, fail_on_missing=False, whitelist=[])
        for target in targets:
            shared_dependencies = set()
            for dependency_path in target.dependencies.used:
                try:
                    relative_dependency_path = dependency_path.relative_to(
                        paths.shared_latex_dir
                    )
                    shared_dependencies.add(str(relative_dependency_path))
                except ValueError:
                    pass
            all_targets_dependencies[targets_name][target.name] = shared_dependencies

    def _relative_to_shared_latex(file_path: str) -> Optional[str]:
        try:
            return str((paths.git_dir / file_path).relative_to(paths.shared_latex_dir))
        except ValueError:
            return None

    try:
        repository = Repo(str(path), search_parent_directories=True)
        index = repository.index
        staged_diffs = index.diff("HEAD")
        unstaged_diffs = index.diff(None)
        untracked_files = repository.untracked_files
        touched = set()
        for diff in chain(staged_diffs, unstaged_diffs):
            for p in [diff.a_path, diff.b_path]:
                relative_to_shared_latex = _relative_to_shared_latex(p)
                if relative_to_shared_latex is not None:
                    touched.add(relative_to_shared_latex)
        for untracked_file in untracked_files:
            relative_to_shared_latex = _relative_to_shared_latex(untracked_file)
            if relative_to_shared_latex is not None:
                touched.add(relative_to_shared_latex)
    except InvalidGitRepositoryError as e:
        raise DeckzException(
            "Could not find the path of the current git working directory. "
            "Are you in one?"
        ) from e

    for targets_name, targets_dependencies in all_targets_dependencies.items():
        touched_targets = []
        for target_name, target_dependencies in targets_dependencies.items():
            if touched.intersection(target_dependencies):
                touched_targets.append(target_name)
        if touched_targets:
            secho(targets_name, fg="green", nl=False)
            secho(f" {' '.join(touched_targets)}", fg="blue")



from typing import List

from deckz.builder import Builder
from deckz.config import get_config
from deckz.paths import Paths
from deckz.targets import Targets


def run(
    paths: Paths,
    build_handout: bool,
    build_presentation: bool,
    build_print: bool,
    target_whitelist: List[str],
) -> None:
    config = get_config(paths)
    targets = Targets(paths=paths, fail_on_missing=True, whitelist=target_whitelist)
    Builder(
        config,
        paths,
        targets,
        build_handout=build_handout,
        build_presentation=build_presentation,
        build_print=build_print,
    )



from logging import getLogger

from deckz.cli import command, deck_path_option, option
from deckz.paths import Paths
from deckz.targets import Targets


@command
@deck_path_option
@option(
    "--dry-run/--do-it",
    default=True,
    help="Print what would be removed instead of removing it.",
)
def clean_latex(deck_path: str, dry_run: bool) -> None:
    """Clean LaTeX files that are not used in `targets*.yml`."""
    logger = getLogger(__name__)
    logger.info("Cleaning unused LaTeX files")
    paths = Paths(deck_path)
    dependencies_dict = Targets(
        paths=paths, fail_on_missing=False, whitelist=[]
    ).get_dependencies()
    for target_name, dependencies in dependencies_dict.items():
        logger.info(f"Processing target {target_name}")
        if dependencies.unused:
            logger.info(
                "%s the following unused dependencies:\n%s",
                "Would remove" if dry_run else "Removing",
                "\n".join(f"  - {str(d)}" for d in sorted(dependencies.unused)),
            )
        else:
            logger.info("All LaTeX files are used, nothing to remove")
        if not dry_run:
            for dependency in dependencies.unused:
                dependency.unlink()



from logging import getLogger
from pathlib import Path
from subprocess import CalledProcessError, run

from yaml import safe_dump, safe_load

from deckz.cli import command
from deckz.exceptions import DeckzException
from deckz.targets import SECTION_YML_VERSION
from deckz.utils import get_section_config_paths


@command
def migrate_sections() -> None:
    """Find all `section.yml` files and upgrade to newest format."""
    logger = getLogger(__name__)

    new_config_paths = []

    for config_path in get_section_config_paths():
        with config_path.open(encoding="utf8") as fh:
            config = safe_load(fh)
        version = config.get("version", 1)
        if version == SECTION_YML_VERSION:
            continue
        logger.info(
            "Upgrading %s from version %d to %d",
            config_path,
            version,
            SECTION_YML_VERSION,
        )
        if version < 2:
            config_path = _v1_v2(config_path)
        if version < 3:
            config_path = _v2_v3(config_path)
        new_config_paths.append(config_path)

    logger.info("Running prettier on section config files")
    try:
        run(
            ["prettier", "--write", *map(str, new_config_paths)],
            check=True,
            capture_output=True,
        )
    except CalledProcessError as e:
        raise DeckzException(
            "Prettier errored or was not found.\n"
            "Captured stdout\n"
            "---\n"
            "%s\n"
            "Captured stderr\n"
            "---\n%s" % (e.stdout, e.stderr)
        ) from e


def _v1_v2(config_path: Path) -> Path:
    with config_path.open(encoding="utf8") as fh:
        config = safe_load(fh)
    title = config["title"]
    includes = [(i if isinstance(i, dict) else {"path": i}) for i in config["includes"]]
    default_titles = {i["path"]: i["title"] for i in includes if "title" in i}
    flavors = {"standard": [i["path"] for i in includes]}
    with config_path.open(encoding="utf8", mode="w") as fh:
        safe_dump(
            dict(
                version=2, title=title, default_titles=default_titles, flavors=flavors
            ),
            fh,
            allow_unicode=True,
        )
    return config_path


def _v2_v3(config_path: Path) -> Path:
    with config_path.open(encoding="utf8") as fh:
        config = safe_load(fh)
    config["version"] = 3
    new_config_path = config_path.parent / f"{config_path.parent.name}.yml"
    with new_config_path.open(encoding="utf8", mode="w") as fh:
        safe_dump(config, fh, allow_unicode=True)
    if config_path != new_config_path:
        config_path.unlink()
    return new_config_path



from contextlib import ExitStack
from enum import Enum
from filecmp import cmp
from logging import getLogger
from multiprocessing import Pool
from os import unlink
from os.path import join as path_join
from pathlib import Path
from shutil import copyfile, move
from subprocess import CompletedProcess, run
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional

from jinja2 import Environment, FileSystemLoader
from PyPDF2 import PdfFileMerger

from deckz.exceptions import DeckzException
from deckz.paths import Paths
from deckz.targets import Target, Targets


class CompileType(Enum):
    Handout = "handout"
    Presentation = "presentation"
    Print = "print"
    PrintHandout = "print-handout"


class Builder:
    def __init__(
        self,
        config: Dict[str, Any],
        paths: Paths,
        targets: Targets,
        build_presentation: bool,
        build_handout: bool,
        build_print: bool,
    ):
        self._config = config
        self._paths = paths
        self._targets = targets
        self._presentation = build_presentation
        self._handout = build_handout
        self._print = build_print
        self._logger = getLogger(__name__)
        self._build_all()

    def _build_all(self) -> None:
        to_compile = []
        for target in self._targets:
            if self._presentation:
                to_compile.append((target, CompileType.Presentation, True))
            if self._handout:
                to_compile.append((target, CompileType.Handout, True))
            if self._print:
                to_compile.append((target, CompileType.PrintHandout, False))
        to_compile.sort(key=lambda x: (x[0].name, x[1].value))
        n_outputs = (
            int(self._handout) * (len(self._targets) + 1)
            + int(self._presentation) * len(self._targets)
            + int(self._print)
        )
        self._logger.info(
            f"Building {len(to_compile)} PDFs used in {n_outputs} outputs"
        )
        with Pool() as pool:
            completed_processes = pool.starmap(self._build, to_compile)
        print_handout_ok = (
            sum(
                c.returncode
                for c, (_, t, _) in zip(completed_processes, to_compile)
                if t is CompileType.PrintHandout
            )
            == 0
        )
        handout_ok = (
            sum(
                c.returncode
                for c, (_, t, _) in zip(completed_processes, to_compile)
                if t is CompileType.Handout
            )
            == 0
        )
        if self._print:
            if print_handout_ok:
                self._logger.info(
                    f"Formatting {len(self._targets)} PDFs into a printable output"
                )
                completed_processes.append(self._build(None, CompileType.Print, True))
            else:
                self._logger.warning(
                    "Preparatory compilations failed. Skipping print output"
                )
        if self._handout:
            if handout_ok:
                self._logger.info(
                    f"Formatting {len(self._targets)} PDFs into a handout"
                )
                self._merge_pdfs(CompileType.Handout, True)
            else:
                self._logger.warning(
                    "Preparatory compilations failed. Skipping handout output"
                )
        for completed_process, (target, compile_type, _) in zip(
            completed_processes, to_compile
        ):
            if target is not None:
                compilation = f"{target.name}/{compile_type.value}"
            else:
                compilation = compile_type.value
            if completed_process.returncode != 0:
                self._logger.warning("Compilation %s errored", compilation)
                self._logger.warning(
                    "Captured %s stderr\n%s", compilation, completed_process.stderr
                )
                self._logger.warning(
                    "Captured %s stdout\n%s", compilation, completed_process.stdout
                )

    def _get_filename(self, target: Optional[Target], compile_type: CompileType) -> str:
        name = self._config["deck_acronym"]
        if target is not None:
            name += f"-{target.name}"
        name += f"-{compile_type.value}"
        return name.lower()

    def _merge_pdfs(self, compile_type: CompileType, copy_result: bool = True) -> None:
        build_dir = self._setup_build_dir(None, compile_type)
        filename = self._get_filename(None, compile_type)
        build_pdf_path = build_dir / f"{filename}.pdf"
        output_pdf_path = self._paths.pdf_dir / f"{filename}.pdf"
        input_pdf_paths = [
            (
                self._paths.build_dir
                / target.name
                / CompileType.Handout.value
                / self._get_filename(target, CompileType.Handout)
            ).with_suffix(".pdf")
            for target in self._targets
        ]
        with ExitStack() as stack, build_pdf_path.open("wb") as fh:
            merger = PdfFileMerger()
            input_files = [stack.enter_context(p.open("rb")) for p in input_pdf_paths]
            for input_file in input_files:
                merger.append(input_file)
            merger.write(fh)
        if copy_result:
            self._paths.pdf_dir.mkdir(parents=True, exist_ok=True)
            copyfile(build_pdf_path, output_pdf_path)

    def _build(
        self,
        target: Optional[Target],
        compile_type: CompileType,
        copy_result: bool = True,
    ) -> CompletedProcess:
        build_dir = self._setup_build_dir(target, compile_type)
        filename = self._get_filename(target, compile_type)
        latex_path = build_dir / f"{filename}.tex"
        build_pdf_path = latex_path.with_suffix(".pdf")
        output_pdf_path = self._paths.pdf_dir / f"{filename}.pdf"
        if compile_type is CompileType.Print:
            self._write_print_latex(self._targets, latex_path)
        else:
            self._write_main_latex(target, compile_type, latex_path)
            self._link_dependencies(target, build_dir)

        completed_process = self._compile(
            latex_path=latex_path.relative_to(build_dir), build_dir=build_dir,
        )
        if copy_result and completed_process.returncode == 0:
            self._paths.pdf_dir.mkdir(parents=True, exist_ok=True)
            copyfile(build_pdf_path, output_pdf_path)
        return completed_process

    def _setup_build_dir(
        self, target: Optional[Target], compile_type: CompileType
    ) -> Path:
        target_build_dir = self._paths.build_dir
        if target is not None:
            target_build_dir /= target.name
        target_build_dir /= compile_type.value
        target_build_dir.mkdir(parents=True, exist_ok=True)
        for item in self._paths.shared_dir.iterdir():
            self._setup_link(target_build_dir / item.name, item)
        return target_build_dir

    def _write_main_latex(
        self, target: Target, compile_type: CompileType, output_path: Path
    ) -> None:
        self._write_latex(
            template_path=self._paths.jinja2_main_template,
            output_path=output_path,
            config=self._config,
            target=target,
            handout=compile_type in [CompileType.Handout, CompileType.PrintHandout],
            print=compile_type is CompileType.PrintHandout,
        )

    def _write_print_latex(self, targets: Targets, output_path: Path) -> None:
        self._write_latex(
            template_path=self._paths.jinja2_print_template,
            output_path=output_path,
            pdf_paths=[
                "../%s/%s/%s"
                % (
                    target.name,
                    CompileType.PrintHandout.value,
                    self._get_filename(target, CompileType.PrintHandout),
                )
                for target in targets
            ],
            format="1x2",
        )

    def _write_latex(
        self, *, template_path: Path, output_path: Path, **template_kwargs: Any
    ) -> None:
        template = self._env.get_template(str(template_path.name))
        try:
            with NamedTemporaryFile("w", encoding="utf8", delete=False) as fh:
                fh.write(template.render(**template_kwargs))
                fh.write("\n")
            if not output_path.exists() or not cmp(fh.name, str(output_path)):
                move(fh.name, output_path)
        finally:
            try:
                unlink(fh.name)
            except FileNotFoundError:
                pass

    @property
    def _env(self) -> Environment:
        if not hasattr(self, "__env"):
            self.__env = Environment(
                loader=FileSystemLoader(searchpath=self._paths.jinja2_dir),
                block_start_string=r"\BLOCK{",
                block_end_string="}",
                variable_start_string=r"\VAR{",
                variable_end_string="}",
                comment_start_string=r"\#{",
                comment_end_string="}",
                line_statement_prefix="%%",
                line_comment_prefix="%#",
                trim_blocks=True,
                autoescape=False,
            )
            self.__env.filters["camelcase"] = self._to_camel_case
            self.__env.filters["path_join"] = lambda paths: path_join(*paths)
        return self.__env

    def _link_dependencies(self, target: Target, target_build_dir: Path) -> None:
        for dependency in target.dependencies.used:
            try:
                link_dir = (
                    target_build_dir
                    / dependency.relative_to(self._paths.shared_latex_dir).parent
                )
            except ValueError:
                link_dir = (
                    target_build_dir
                    / dependency.relative_to(
                        self._paths.current_dir / target.name
                    ).parent
                )
            link_dir.mkdir(parents=True, exist_ok=True)
            self._setup_link(link_dir / dependency.name, dependency)

    def _compile(self, latex_path: Path, build_dir: Path) -> CompletedProcess:
        command = [
            "latexmk",
            "-pdflatex=xelatex -shell-escape -interaction=nonstopmode %O %S",
            "-dvi-",
            "-ps-",
            "-pdf",
        ]
        command.append(str(latex_path))
        return run(command, cwd=build_dir, capture_output=True, encoding="utf8")

    def _setup_link(self, source: Path, target: Path) -> None:
        if not target.exists():
            raise DeckzException(
                f"{target} could not be found. Please make sure it exists before "
                "proceeding."
            )
        target = target.resolve()
        if source.is_symlink():
            if source.resolve().samefile(target):
                return
            raise DeckzException(
                f"{source} already exists in the build directory and does not point to "
                f"{target}. Please clean the build directory."
            )
        elif source.exists():
            raise DeckzException(
                f"{source} already exists in the build directory. Please clean the "
                "build directory."
            )
        source.parent.mkdir(parents=True, exist_ok=True)
        source.symlink_to(target)

    def _to_camel_case(self, string: str) -> str:
        return "".join(substring.capitalize() or "_" for substring in string.split("_"))



from itertools import chain
from pathlib import Path
from typing import FrozenSet, Optional

from git import Repo
from git.exc import InvalidGitRepositoryError
from yaml import safe_load

from deckz.exceptions import DeckzException


def get_section_config_paths() -> FrozenSet[Path]:
    git_dir = get_git_dir(Path("."))
    v1_ymls = git_dir.glob("**/section.yml")
    all_ymls = git_dir.glob("**/*.yml")
    vx_ymls = []
    for yml in all_ymls:
        with yml.open(encoding="utf8") as fh:
            content = safe_load(fh)
            if not isinstance(content, dict):
                continue
            if {"title", "version", "flavors"}.issubset(content):
                vx_ymls.append(yml)
    return frozenset(chain(v1_ymls, vx_ymls))


def get_git_dir(path: Path) -> Optional[Path]:
    try:
        repository = Repo(str(path), search_parent_directories=True)
    except InvalidGitRepositoryError as e:
        raise DeckzException(
            "Could not find the path of the current git working directory. "
            "Are you in one?"
        ) from e
    return Path(repository.git.rev_parse("--show-toplevel")).resolve()



from collections import ChainMap, OrderedDict
from logging import getLogger
from pathlib import Path
from shutil import copy as shutil_copy
from typing import Any, Dict, Optional

from yaml import safe_load as yaml_safe_load

from deckz.exceptions import DeckzException
from deckz.paths import Paths


_logger = getLogger(__name__)


def get_config(paths: Paths) -> Dict[str, Any]:
    return OrderedDict(
        (k, v)
        for k, v in sorted(
            ChainMap(
                *[
                    _get_or_create_config(config_path, template_path)
                    for config_path, template_path in [
                        (paths.session_config, None),
                        (paths.deck_config, paths.template_deck_config),
                        (paths.company_config, paths.template_company_config),
                        (paths.user_config, paths.template_user_config),
                        (paths.global_config, paths.template_global_config),
                    ]
                ],
            ).items()
        )
    )


def _get_or_create_config(
    config_path: Path, template_path: Optional[Path]
) -> Dict[str, Any]:
    if not config_path.is_file():
        if template_path:
            if template_path.is_file():
                shutil_copy(str(template_path), str(config_path), follow_symlinks=True)
                raise DeckzException(
                    f"{config_path} was not found, copied {template_path} there. "
                    "Please edit it."
                )
            else:
                raise DeckzException(
                    f"Neither {config_path} nor {template_path} were found. "
                    "Please create both."
                )
        else:
            return {}

    with open(config_path, "r", encoding="utf8") as fh:
        return yaml_safe_load(fh)



from typing import List

from deckz.cli import (
    command,
    compile_type_options,
    deck_path_option,
    target_whitelist_argument,
)
from deckz.paths import Paths
from deckz.runner import run as runner_run


@command
@target_whitelist_argument
@deck_path_option
@compile_type_options(
    default_handout=True, default_presentation=True, default_print=True
)
def run(
    deck_path: str,
    target_whitelist: List[str],
    build_handout: bool,
    build_presentation: bool,
    build_print: bool,
) -> None:
    """Compile main targets."""
    paths = Paths(deck_path)
    runner_run(
        paths=paths,
        build_handout=build_handout,
        build_presentation=build_presentation,
        build_print=build_print,
        target_whitelist=target_whitelist,
    )



from logging import getLogger
from pathlib import Path
from threading import Thread
from time import time
from typing import List, Optional

from watchdog.events import FileSystemEvent
from watchdog.observers import Observer

from deckz.cli import (
    command,
    compile_type_options,
    deck_path_option,
    option,
    target_whitelist_argument,
)
from deckz.exceptions import DeckzException
from deckz.paths import Paths
from deckz.runner import run


@command
@target_whitelist_argument
@deck_path_option
@compile_type_options(
    default_handout=False, default_presentation=True, default_print=False
)
@option(
    "--minimum-delay", default=5, type=int, help="Minimum delay before recompiling.",
)
def watch(
    minimum_delay: int,
    deck_path: str,
    build_handout: bool,
    build_presentation: bool,
    build_print: bool,
    target_whitelist: List[str],
) -> None:
    """Compile on change."""

    class LatexCompilerEventHandler:
        def __init__(
            self,
            minimum_delay: int,
            paths: Paths,
            build_handout: bool,
            build_presentation: bool,
            build_print: bool,
            target_whitelist: List[str],
        ):
            self._minimum_delay = minimum_delay
            self._last_compile = 0.0
            self._paths = paths
            self._build_handout = build_handout
            self._build_presentation = build_presentation
            self._build_print = build_print
            self._target_whitelist = target_whitelist
            self._worker: Optional[Thread] = None

        def work(self) -> None:
            try:
                self._compiling = True
                logger.info("Detected changes, starting a new build")
                try:
                    run(
                        paths=self._paths,
                        build_handout=self._build_handout,
                        build_presentation=self._build_presentation,
                        build_print=self._build_print,
                        target_whitelist=self._target_whitelist,
                    )
                    logger.info("Build finished")
                except Exception as e:
                    logger.critical("Build failed. Error: %s", str(e))
            finally:
                self._compiling = False

        def dispatch(self, event: FileSystemEvent) -> None:
            for d in [paths.build_dir, paths.pdf_dir]:
                if d in Path(event.src_path).parents:
                    return
            current_time = time()
            if self._last_compile + self._minimum_delay > current_time:
                return
            elif self._worker is not None and self._worker.is_alive():
                logger.info("Still on last build, not starting a new build")
                return
            else:
                self._last_compile = current_time
                self._worker = Thread(target=self.work)
                self._worker.start()

    logger = getLogger(__name__)
    logger.info("Watching current and shared directories")
    paths = Paths(deck_path)
    observer = Observer()
    event_handler = LatexCompilerEventHandler(
        minimum_delay,
        paths=paths,
        build_handout=build_handout,
        build_print=build_print,
        build_presentation=build_presentation,
        target_whitelist=target_whitelist,
    )
    paths_to_watch = [
        (p, False) for p in paths.shared_dir.glob("**/*") if p.resolve().is_dir()
    ]
    paths_to_watch.append((paths.jinja2_dir, True))
    paths_to_watch.append((paths.current_dir, True))
    for path, recursive in paths_to_watch:
        observer.schedule(event_handler, str(path.resolve()), recursive=recursive)
    observer.start()
    try:
        while observer.isAlive():
            observer.join(1)
    except KeyboardInterrupt:
        observer.stop()
        observer.join()
        logger.info("Stopped watching")
    else:
        observer.join()
        raise DeckzException("Stopped watching abnormally")



from collections import OrderedDict
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Tuple

from yaml import safe_load as yaml_safe_load

from deckz.exceptions import DeckzException
from deckz.paths import Paths


_logger = getLogger(__name__)


SECTION_YML_VERSION = 3


class Section:
    def __init__(self, title: Optional[str]):
        self.title = title
        self.inputs: List[Tuple[str, Optional[str]]] = []

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"title={repr(self.title)},"
            f"inputs={repr(self.inputs)})"
        )


class Dependencies:
    def __init__(self) -> None:
        self.used: Set[Path] = set()
        self.missing: Set[str] = set()
        self.unused: Set[Path] = set()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"used={repr(self.used)},"
            f"missing={repr(self.missing)},"
            f"unused={repr(self.unused)})"
        )

    @staticmethod
    def merge(*dependencies_list: "Dependencies") -> "Dependencies":
        dependencies = Dependencies()
        for ds in dependencies_list:
            dependencies.used |= ds.used
            dependencies.missing |= ds.missing
            dependencies.unused |= ds.unused - dependencies.used
        return dependencies

    def merge_dicts(
        *dependencies_dicts: Dict[str, "Dependencies"]
    ) -> Dict[str, "Dependencies"]:
        keys = set.union(*(set(d) for d in dependencies_dicts))
        merged_dict = {}
        for key in keys:
            merged_dict[key] = Dependencies.merge(
                *(d[key] for d in dependencies_dicts if key in d)
            )
        return merged_dict


class Target:
    def __init__(self, data: Dict[str, Any], paths: Paths):
        self._paths = paths
        self.name = data["name"]
        local_latex_dir = paths.current_dir / self.name
        self.title = data["title"]
        self.dependencies = Dependencies()
        self.dependencies.unused.update(local_latex_dir.glob("**/*.tex"))
        self.sections = []
        for section_config in data["sections"]:
            if not isinstance(section_config, dict):
                section_config = dict(path=section_config)
            section_path = section_config["path"]
            local_dir = local_latex_dir / section_path
            local_file = local_latex_dir / f"{section_path}.tex"
            shared_dir = paths.shared_latex_dir / section_path
            shared_file = paths.shared_latex_dir / f"{section_path}.tex"
            if local_dir.exists() and local_dir.resolve().is_dir():
                section, dependencies = self._parse_section_dir(
                    local_dir, local_latex_dir, section_config
                )
            elif local_file.exists() and local_file.resolve().is_file():
                section, dependencies = self._parse_section_file(
                    local_file, local_latex_dir, section_config
                )
            elif shared_dir.exists() and shared_dir.resolve().is_dir():
                section, dependencies = self._parse_section_dir(
                    shared_dir, paths.shared_latex_dir, section_config
                )
            elif shared_file.exists() and shared_file.resolve().is_file():
                section, dependencies = self._parse_section_file(
                    shared_file, paths.shared_latex_dir, section_config
                )
            else:
                dependencies = Dependencies()
                dependencies.missing.add(section_path)
            self.dependencies = Dependencies.merge(self.dependencies, dependencies)
            self.sections.append(section)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"name={repr(self.name)},"
            f"title={repr(self.title)},"
            f"dependencies={repr(self.dependencies)},"
            f"sections={repr(self.sections)})"
        )

    def _parse_section_dir(
        self, section_dir: Path, latex_dir: Path, custom_config: Dict[str, Any]
    ) -> Tuple[Section, Dependencies]:
        section_config_path = section_dir / f"{section_dir.name}.yml"
        with open(section_config_path, encoding="utf8") as fh:
            section_config = yaml_safe_load(fh)
        title = (
            custom_config["title"]
            if "title" in custom_config
            else section_config["title"]
        )
        if "flavor" not in custom_config:
            raise DeckzException(
                f"Mandatory flavor not specified in {section_dir.name} configuration "
                "of targets.yml."
            )
        flavor_name = custom_config["flavor"]
        if "flavors" not in section_config:
            raise DeckzException(
                f"Mandatory dictionary `flavors` not found in {section_config_path}."
            )
        flavors = section_config["flavors"]
        if flavor_name not in flavors:
            flavors_string = ", ".join("'%s'" % f for f in flavors)
            raise DeckzException(
                f"'{flavor_name}' not amongst available flavors: {flavors_string} "
                f"of {section_config_path}."
            )
        flavor = flavors[flavor_name]
        section = Section(title)
        dependencies = Dependencies()
        default_titles = section_config.get("default_titles")
        for item in flavor:
            if isinstance(item, str):
                filename = item
                if default_titles is not None:
                    title = default_titles.get(filename)
                else:
                    title = None
            else:
                filename, title = next(iter(item.items()))
            if "excludes" in custom_config and filename in custom_config["excludes"]:
                continue
            local_path = (section_dir / filename).with_suffix(".tex")
            shared_path = (self._paths.shared_latex_dir / filename).with_suffix(".tex")
            if local_path.exists():
                section.inputs.append(
                    (f"{section_dir.relative_to(latex_dir)}/{filename}", title)
                )
                dependencies.used.add(local_path.resolve())
            elif shared_path.exists():
                section.inputs.append((filename, title))
                dependencies.used.add(shared_path.resolve())
            else:
                dependencies.missing.add(filename)
        return section, dependencies

    def _parse_section_file(
        self, section_file: Path, latex_dir: Path, config: Dict[str, Any]
    ) -> Tuple[Section, Dependencies]:
        config_file = section_file.with_suffix(".yml")
        if "title" in config:
            title = config["title"]
        elif config_file.exists():
            with config_file.open(encoding="utf8") as fh:
                default_config = yaml_safe_load(fh)
            title = default_config["title"]
        else:
            title = None
        section = Section(title)
        section.inputs.append(
            (f"{section_file.relative_to(latex_dir).with_suffix('')}", None)
        )
        dependencies = Dependencies()
        dependencies.used.add(section_file.resolve())
        return section, dependencies


class Targets(Iterable[Target]):
    def __init__(
        self, paths: Paths, fail_on_missing: bool, whitelist: List[str]
    ) -> None:
        self._paths = paths
        if not paths.targets.exists():
            if fail_on_missing:
                raise DeckzException(f"Could not find {paths.targets}.")
            else:
                self.targets = []
        with paths.targets.open("r", encoding="utf8") as fh:
            targets = [
                Target(data=target, paths=paths) for target in yaml_safe_load(fh)
            ]
        missing_dependencies = {
            target.name: target.dependencies.missing
            for target in targets
            if target.dependencies.missing
        }
        if missing_dependencies:
            raise DeckzException(
                "Could not find the following dependencies:\n%s"
                % "\n".join(
                    "  - %s:\n%s" % (k, "\n".join(f"    - {p}" for p in v))
                    for k, v in missing_dependencies.items()
                ),
            )
        target_names = set(target.name for target in targets)
        whiteset = set(whitelist)
        unmatched = whiteset - target_names
        if unmatched:
            raise DeckzException(
                "Could not find the following targets:\n%s"
                % "\n".join("  - %s" % name for name in unmatched)
            )
        self.targets = [
            target for target in targets if not whiteset or target.name in whiteset
        ]

    def get_dependencies(self) -> Dict[str, Dependencies]:
        return OrderedDict((t.name, t.dependencies) for t in self.targets)

    def __iter__(self) -> Iterator[Target]:
        return iter(self.targets)

    def __len__(self) -> int:
        return len(self.targets)

    def __repr__(self) -> str:
        return f"Targets(targets={repr(self.targets)}"



from logging import getLogger
from pathlib import Path

from appdirs import user_config_dir

from deckz import app_name
from deckz.exceptions import DeckzException
from deckz.utils import get_git_dir


_logger = getLogger(__name__)


class GlobalPaths:
    def __init__(self, current_dir: str) -> None:
        self.current_dir = Path(current_dir).resolve()
        self.shared_dir = self.git_dir / "shared"
        self.shared_img_dir = self.shared_dir / "img"
        self.shared_code_dir = self.shared_dir / "code"
        self.shared_latex_dir = self.shared_dir / "latex"
        self.templates_dir = self.git_dir / "templates"
        self.yml_templates_dir = self.templates_dir / "yml"
        self.template_targets = self.yml_templates_dir / "targets.yml"
        self.template_global_config = self.yml_templates_dir / "global-config.yml"
        self.template_user_config = self.yml_templates_dir / "user-config.yml"
        self.template_company_config = self.yml_templates_dir / "company-config.yml"
        self.template_deck_config = self.yml_templates_dir / "deck-config.yml"
        self.jinja2_dir = self.templates_dir / "jinja2"
        self.jinja2_main_template = self.jinja2_dir / "main.tex"
        self.jinja2_print_template = self.jinja2_dir / "print.tex"
        self.user_config_dir = Path(user_config_dir(app_name))
        self.global_config = self.git_dir / "global-config.yml"
        self.gdrive_secrets = self.user_config_dir / "gdrive-secrets.json"
        self.gdrive_credentials = self.user_config_dir / "gdrive-credentials.pickle"
        self.user_config = self.user_config_dir / "user-config.yml"

    @property
    def git_dir(self) -> Path:
        if not hasattr(self, "_git_dir"):
            self._git_dir = get_git_dir(self.current_dir)
        return self._git_dir


class Paths(GlobalPaths):
    def __init__(self, current_dir: str) -> None:
        super().__init__(current_dir)

        if not self.current_dir.relative_to(self.git_dir).match("*/*"):
            raise DeckzException(
                f"Not deep enough from root {self.git_dir}. "
                "Please follow the directory hierarchy root > company > deck and "
                "invoke this tool from the deck directory."
            )

        self.build_dir = self.current_dir / "build"
        self.pdf_dir = self.current_dir / "pdf"
        self.company_config = (
            self.git_dir
            / self.current_dir.relative_to(self.git_dir).parts[0]
            / "company-config.yml"
        )
        self.deck_config = self.current_dir / "deck-config.yml"
        self.session_config = self.current_dir / "session-config.yml"
        self.targets = self.current_dir / "targets.yml"

        self.user_config_dir.mkdir(parents=True, exist_ok=True)



from logging import getLogger

from deckz.cli import (
    command,
    deck_path_option,
)
from deckz.paths import Paths
from deckz.uploader import Uploader


_logger = getLogger(__name__)


@command
@deck_path_option
def upload(deck_path: str) -> None:
    """Upload pdfs to Google Drive."""
    paths = Paths(deck_path)
    Uploader(paths)



from functools import partial, wraps
from importlib import import_module, invalidate_caches as importlib_invalidate_caches
from logging import INFO
from pkgutil import walk_packages
from typing import Any, Callable, List, Tuple

from click import (
    argument,
    ClickException,
    group,
    option as click_option,
    Path as ClickPath,
)
from coloredlogs import install as coloredlogs_install

from deckz.exceptions import DeckzException
from deckz.paths import Paths
from deckz.targets import Targets


option = partial(click_option, show_default=True)


deck_path_option = option(
    "--deck-path",
    type=ClickPath(exists=True, readable=True, file_okay=False),
    default=".",
    help="Path of the deck.",
)


def presentation_option(
    default: bool = True,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    return option(
        "--presentation/--no-presentation",
        "build_presentation",
        default=default,
        help="Compile the presentation version.",
    )


def handout_option(
    default: bool = True,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    return option(
        "--handout/--no-handout",
        "build_handout",
        default=default,
        help="Compile the handout version.",
    )


def print_option(
    default: bool = True,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    return option(
        "--print/--no-print",
        "build_print",
        default=default,
        help="Compile the printable version.",
    )


def compile_type_options(
    default_presentation: bool = True,
    default_handout: bool = True,
    default_print: bool = True,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    def inner(f: Callable[..., Any]) -> Callable[..., Any]:
        return print_option(default=default_print)(
            presentation_option(default=default_presentation)(
                handout_option(default=default_handout)(f)
            )
        )

    return inner


def _autocomplete_target_whitelist(
    ctx: Any, args: List[str], incomplete: str
) -> List[Tuple[str, str]]:
    try:
        paths = Paths(".")
        targets = Targets(paths=paths, fail_on_missing=False, whitelist=[])
        return [(t.name, t.title) for t in targets if incomplete in t.name]
    except Exception:
        return []


target_whitelist_argument = argument(
    "target_whitelist",
    metavar="targets",
    nargs=-1,
    autocompletion=_autocomplete_target_whitelist,  # type: ignore
)


@group(chain=True)
def cli() -> None:
    coloredlogs_install(
        level=INFO, fmt="%(asctime)s %(name)s %(message)s", datefmt="%H:%M:%S",
    )


def command(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            result = f(*args, **kwargs)
        except DeckzException as e:
            raise ClickException(str(e)) from e
        return result

    return cli.command()(wrapper)


def _import_module_and_submodules(package_name: str) -> None:
    """
    From https://github.com/allenai/allennlp/blob/master/allennlp/common/util.py
    """
    importlib_invalidate_caches()

    module = import_module(package_name)
    path = getattr(module, "__path__", [])
    path_string = "" if not path else path[0]

    for module_finder, name, _ in walk_packages(path):
        if path_string and module_finder.path != path_string:
            continue
        subpackage = f"{package_name}.{name}"
        _import_module_and_submodules(subpackage)


_import_module_and_submodules(__name__)



from logging import getLogger
from shutil import copy as shutil_copy

from deckz.cli import command, deck_path_option
from deckz.paths import Paths


@command
@deck_path_option
def init(deck_path: str) -> None:
    """Create an initial targets.yml."""
    logger = getLogger(__name__)
    paths = Paths(deck_path)
    if paths.targets.exists():
        logger.info(f"Nothing to do: {paths.targets} already exists")

    else:
        logger.info(f"Copying {paths.template_targets} to current directory")
        shutil_copy(str(paths.template_targets), str(paths.current_dir))



from logging import getLogger

from deckz.cli import command, deck_path_option
from deckz.config import get_config
from deckz.paths import Paths


@command
@deck_path_option
def print_config(deck_path: str) -> None:
    """Print the resolved configuration."""
    logger = getLogger(__name__)
    paths = Paths(deck_path)
    config = get_config(paths)
    logger.info(
        "Resolved config as:\n%s",
        "\n".join((f"  - {k}: {v}") for k, v in config.items()),
    )
